﻿--[[
Name: Sink-1.0
Revision: $Rev: 44886 $
Author(s): Rabbit (rabbit.magtheridon@gmail.com
Website: http://rabbit.nihilum.eu
Documentation: http://wiki.wowace.com/index.php/Sink-1.0
SVN: http://svn.wowace.com/wowace/trunk/SinkLib/Sink-1.0
Description: Library that handles chat output.
Dependencies: AceLibrary, AceOO-2.0, SharedMedia-1.0 (optional)
License: GPL v2 or later.
]]

--[[
Copyright (C) 2007 Rabbit

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
]]

-----------------------------------------------------------------------
-- Sink-1.0

local SINK10 = "Sink-1.0"
local SINK10_MINOR = string.match("$Revision: 44886 $", "[0-9]+")

assert(AceLibrary, SINK10 .. " requires AceLibrary.")
if not AceLibrary:IsNewVersion(SINK10, SINK10_MINOR) then return end
if not AceLibrary:HasInstance("AceOO-2.0") then error(SINK10 .. " requires AceOO-2.0") end

local L_DEFAULT = "Default"
local L_ROUTE = "Route output from this addon through %s."
local L_SCT = "Scrolling Combat Text"
local L_MSBT = "MikSBT"
local L_BIGWIGS = "BigWigs"
local L_CHAT = "Chat"
local L_BLIZZARD = "Blizzard FCT"
local L_RW = "Raid Warning"
local L_PARROT = "Parrot"
local L_OUTPUT = "Output"
local L_OUTPUT_DESC = "Where to route the output from this addon."
local L_SCROLL = "Scroll area"
local L_SCROLL_DESC = "Set the scrollarea where messages should appear."
local L_NONE = "None"
local L_NONE_DESC = "Hide all messages from this addon."

local l = GetLocale()
if l == "koKR" then
	L_DEFAULT = "기본"
	L_ROUTE = "%s|1을;를; 통해 이 애드온의 메시지를 출력합니다."
	L_SCT = "Scrolling Combat Text"
	L_MSBT = "MikSBT"
	L_BIGWIGS = "BigWigs"
	L_CHAT = "대화창"
	L_BLIZZARD = "블리자드 FCT"
	L_RW = "공격대 경보"
	L_PARROT = "Parrot"
	L_OUTPUT = "출력"
	L_OUTPUT_DESC = "이 애드온의 출력 메시지를 변경합니다."
	L_SCROLL = "스크롤 영역"
	L_SCROLL_DESC = "메시지를 출력할 스크룰 영역을 설정합니다."
	L_NONE = "없음"
	L_NONE_DESC = "이 애드온의 모든 메시지를 숨김니다."
elseif l == "frFR" then
	L_DEFAULT = "Par défaut"
	L_ROUTE = "Transmet la sortie de cet addon via %s."
	-- L_SCT = "Scrolling Combat Text"
	-- L_MSBT = "MikSBT"
	-- L_BIGWIGS = "BigWigs"
	L_CHAT = "Fenêtre de discussion"
	L_BLIZZARD = "Texte de Combat Blizzard"
	L_RW = "Alerte Raid"
	-- L_PARROT = "Parrot"
	L_OUTPUT = "Sortie"
	L_OUTPUT_DESC = "Destination de la sortie de cet addon."
	L_SCROLL = "Zone de défilement"
	L_SCROLL_DESC = "Définie la zone de défilement où les messages doivent apparaitre."
	L_NONE = "Aucun"
	L_NONE_DESC = "Masque tous les messages de cet addon."
elseif l == "deDE" then
	L_DEFAULT = "Voreinstellung"
	L_ROUTE = "Schickt die Meldungen dieses Addons an %s."
	L_SCT = "Scrolling Combat Text(SCT)"
	L_MSBT = "MikSBT"
	L_BIGWIGS = "BigWigs"
	L_CHAT = "den Chat"
	L_BLIZZARD = "Blizzard's schwebenden Kampftext"
	L_RW = "die Schlachtzug Warnung"
	L_PARROT = "Parrot"
	L_OUTPUT = "Ausgabe"
	L_OUTPUT_DESC = "Wohin die Meldungen des Addons gesendet werden soll."
	L_SCROLL = "Scroll Area"
	L_SCROLL_DESC = "Setzt die Scroll Area, wo die Meldungen erscheinen sollen."
	L_NONE = "Nirgends"
	L_NONE_DESC = "Versteckt alle Meldungen von diesem Addon."
elseif l == "zhCN" then
	-- etc
end

local SML = AceLibrary:HasInstance("SharedMedia-1.0") and AceLibrary("SharedMedia-1.0") or nil

local AceOO = AceLibrary("AceOO-2.0")
local sink = AceOO.Mixin {
	"Pour",
	"RegisterSink",
}

local _G = getfenv(0)

-- Thanks to Antiarc and his Soar-1.0 library for most of the 'meat' of the
-- sink-specific functions.

local function parrot(addon, text, r, g, b, font, size, outline, sticky, loc, icon)
	local location = sink.storageForAddon[addon] and sink.storageForAddon[addon].sink10ScrollArea or "Notification"
	Parrot:ShowMessage(text, location, sticky, r, g, b, font, size, outline, icon)
end

local sct_color = {}
local function sct(addon, text, r, g, b, font, size, outline, sticky)
	sct_color.r, sct_color.g, sct_color.b = r, g, b
	local loc = sink.storageForAddon[addon] and sink.storageForAddon[addon].sink10ScrollArea or "Messages"
	local location = (loc == "Incoming" and SCT.FRAME1) or (location == "Outgoing" and SCT.FRAME2) or SCT.MSG
	SCT:DisplayCustomEvent(text, sct_color, sticky, location)
end

local msbt_registered_fonts = {}
local msbt_outlines = {["NORMAL"] = 1, ["OUTLINE"] = 2, ["THICKOUTLINE"] = 3}
local function msbt(addon, text, r, g, b, font, size, outline, sticky)
	if font and SML and not msbt_registered_fonts[font] then
		MikSBT.RegisterFont(font, SML:Fetch("font", font))
		msbt_registered_fonts[font] = true
	end
	local location = sink.storageForAddon[addon] and sink.storageForAddon[addon].sink10ScrollArea or MikSBT.DISPLAYTYPE_NOTIFICATION
	MikSBT.DisplayMessage(text, location, sticky, r * 255, g * 255, b * 255, size, font, msbt_outlines[outline] or 1)
end

local function blizzard(addon, text, r, g, b, font, size, outline, sticky)
	if tostring(SHOW_COMBAT_TEXT) ~= "0" then
		CombatText_AddMessage(text, CombatText_StandardScroll, r, g, b, sticky and "crit" or nil, false)
	else
		UIErrorsFrame:AddMessage(text, r, g, b, 1, UIERRORS_HOLD_TIME)
	end
end

local function chat(addon, text, r, g, b)
	DEFAULT_CHAT_FRAME:AddMessage(text, r, g, b)
end
local function rw(addon, text, r, g, b)
	RaidWarningFrame:AddMessage(text, r, g, b, 1, UIERRORS_HOLD_TIME)
end

local bwMsg = nil
local function bw(addon, text, r, g, b)
	if not bwMsg and BigWigs and BigWigs:HasModule("Messages") then
		bwMsg = BigWigs:GetModule("Messages")
	end
	if not bwMsg then
		chat(addon, text, r, g, b)
		return
	end
	bwMsg:Print(text, r, g, b)
end

local function noop() --[[ noop! ]] end

local handlers = {
	Parrot = parrot,
	SCT = sct,
	MikSBT = msbt,
	BigWigs = bw,
	-- _G.ChatFrame, Blizzard, RaidWarning and None are all nil on a default install
	ChatFrame = chat,
	Blizzard = blizzard,
	RaidWarning = rw,
	None = noop,
}

function sink:Pour(text, r, g, b, ...)
	local func = sink.override and handlers[sink.override] or nil
	if not func and sink.storageForAddon[self] then
		func = handlers[sink.storageForAddon[self].sink10OutputSink]
	end
	if not func then
		for name, f in pairs(handlers) do
			if _G[name] then
				func = f
				break
			end
		end
	end
	if not func then func = chat end
	func(self, text, r or 1, g or 1, b or 1, ...)
end

do
	local options = {}
	-- Maybe we want to hide them instead of disable
	local function shouldDisableSCT()
		return not _G.SCT
	end
	local function shouldDisableMSBT()
		return not _G.MikSBT
	end
	local function shouldDisableParrot()
		return not _G.Parrot
	end
	local function shouldDisableBigWigs()
		return not _G.BigWigs
	end
	local function shouldDisableFCT()
		return not SHOW_COMBAT_TEXT or tostring(SHOW_COMBAT_TEXT) == "0"
	end

	local sinks = {
		SCT = {
			type = "toggle",
			name = L_SCT,
			desc = L_ROUTE:format(L_SCT),
			isRadio = true,
			hidden = shouldDisableSCT,
		},
		MikSBT = {
			type = "toggle",
			name = L_MSBT,
			desc = L_ROUTE:format(L_MSBT),
			hidden = shouldDisableMSBT,
		},
		Parrot = {
			type = "toggle",
			name = L_PARROT,
			desc = L_ROUTE:format(L_PARROT),
			isRadio = true,
			hidden = shouldDisableParrot,
		},
		BigWigs = {
			type = "toggle",
			name = L_BIGWIGS,
			desc = L_ROUTE:format(L_BIGWIGS),
			isRadio = true,
			hidden = shouldDisableBigWigs,
		},
		Blizzard = {
			type = "toggle",
			name = L_BLIZZARD,
			desc = L_ROUTE:format(L_BLIZZARD),
			isRadio = true,
			hidden = shouldDisableFCT,
		},
		RaidWarning = {
			type = "toggle",
			name = L_RW,
			desc = L_ROUTE:format(L_RW),
			isRadio = true,
		},
		ChatFrame = {
			type = "toggle",
			name = L_CHAT,
			desc = L_ROUTE:format(L_CHAT),
			isRadio = true,
		},
		None = {
			type = "toggle",
			name = L_NONE,
			desc = L_NONE_DESC,
			isRadio = true,
		},
	}
	local sctFrames = {"Incoming", "Outgoing", "Messages"}
	local emptyTable = {}

	function sink:RegisterSink(shortName, name, desc, func)
		sink:argCheck(shortName, 2, "string")
		sink:argCheck(name, 3, "string")
		sink:argCheck(desc, 4, "string", "nil")
		sink:argCheck(func, 5, "string", "function")
		if sinks[shortName] or handlers[shortName] then
			sink:error("There's already a sink by the short name %q.", shortName)
		end
		if _G[shortName] then
			sink:error("Because of a limitation in the implementation, it's currently not possible to register a sink where its shortName also appears in the global scope, _G[%q] is already a %q.", shortName, type(_G[shortName]))
		end
		sinks[shortName] = {
			type = "toggle",
			name = name,
			desc = desc or L_ROUTE:format(name),
			isRadio = true,
		}
		if type(func) == "function" then
			handlers[shortName] = func
		else
			handlers[shortName] = function(...)
				self[func](self, ...)
			end
		end
	end

	function sink:GetAceOptionsDataTable(addon)
		if not options[addon] then
			options[addon] = {
				output = {
					type = "group",
					name = L_OUTPUT,
					desc = L_OUTPUT_DESC,
					pass = true,
					get = function(key)
						if not sink.storageForAddon[addon] then
							return nil
						end
						if tostring(key) == "nil" then
							-- Means AceConsole wants to list the output option,
							-- so we should show which sink is currently used.
							return sink.storageForAddon[addon].sink10OutputSink or L_DEFAULT
						end
						if key == "ScrollArea" then
							return sink.storageForAddon[addon].sink10ScrollArea
						else
							return sink.storageForAddon[addon].sink10OutputSink and sink.storageForAddon[addon].sink10OutputSink == key or nil
						end
					end,
					set = function(key, value)
						if not sink.storageForAddon[addon] then return end
						if key == "ScrollArea" then
							sink.storageForAddon[addon].sink10ScrollArea = value
						elseif value then
							if key == "Parrot" then
								options[addon].output.args.ScrollArea.validate = Parrot:GetScrollAreasValidate()
							elseif key == "MikSBT" then
								options[addon].output.args.ScrollArea.validate = MikSBT.GetScrollAreaList()
							elseif key == "SCT" then
								options[addon].output.args.ScrollArea.validate = sctFrames
							else
								sink.storageForAddon[addon].sink10ScrollArea = nil
								options[addon].output.args.ScrollArea.validate = emptyTable
							end
							sink.storageForAddon[addon].sink10OutputSink = key
						end
					end,
					args = sinks,
				},
			}
			options[addon].output.args.ScrollArea = {
				type = "text",
				name = L_SCROLL,
				desc = L_SCROLL_DESC,
				validate = emptyTable,
				order = -1,
				disabled = function()
					if not sink.storageForAddon[addon] then return true end
					local s = sink.storageForAddon[addon].sink10OutputSink
					if not s then return true end
					return s ~= "Parrot" and s ~= "MikSBT" and s ~= "SCT"
				end,
			}
		end
		return options[addon]
	end
end

do
	local prioritizedStorageTypes = {"profile", "char", "realm", "server", "class", "account"}
	function sink:OnEmbedEnable(addon, name)
		local found = addon.db and true or nil
		if not found and addon.class and addon.class.mixins then
			for mixin in pairs(addon.class.mixins) do
				if tostring(mixin) == "AceDB-2.0" then
					found = true
					break
				end
			end
		end
		if found then
			for i, v in ipairs(prioritizedStorageTypes) do
				if addon.db[v] then
					sink.storageForAddon[addon] = addon.db[v]
					break
				end
			end
		end
	end
end

-- Sets a sink override for -all- addons, librarywide.
function sink:SetSinkOverride(sink)
	sink:argCheck(sink, 2, "string", "nil")
	if not handlers[sink] then
		sink:error("There's no %q sink.", sink)
	end
	sink.override = sink
end

local function activate(self, oldLib, oldDeactivate)
	sink = self
	if oldLib then
		self.storageForAddon = oldLib.storageForAddon
		self.override = oldLib.override
	end
	if not self.storageForAddon then self.storageForAddon = {} end
	if oldLib and oldDeactivate then
		oldDeactivate(oldLib)
	end
end

AceLibrary:Register(sink, SINK10, SINK10_MINOR, activate)

